/**
 * 
 */
package sim.geometry;

import java.io.BufferedWriter;
import java.io.IOException;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.graphics.SPrimitive;
import sim.math.SVector3d;
import sim.math.SVectorUV;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SSphericalCapGeometry</b> reprsente la gomtrie d'une calotte sphrique qui correspond  une tranche d'une sphre.
 * Cette tranche possde un rayon selon la gomtrie d'un disque et une courbure selon la gomtrie d'une sphre.
 * Si l'on coupe une sphre en deux morceaux identiques, la calotte sphrique corespond  une demi-sphre.
 * 
 * <p>Une calotte sphrique tant dfini  l'aide d'un rayon de courbure respecte la convension de signe des dioptres sphriques suivante :
 * <ul>- Courbure positive (R > 0) : Courbure convexe par rapport  la surface du disque (dans le sens de la normale  la surface du disque).</ul>
 * <ul>- Courbure ngative (R < 0) : Courbure concave par rapport  la surface du disque (dans le sens oppos de la normale  la surface du disque).</ul>
 * </p>
 * 
 * @author Simon Vzina
 * @since 2015-11-08
 * @version 2017-12-20 (version labo  Le ray tracer v2.1)
 */
public class SSphericalCapGeometry extends SDiskGeometry {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { SKeyWordDecoder.KW_CURVATURE };
  
  /**
   * La constante <b>DEFAULT_CURVATURE</b> correspond au rayon de courbure par dfaut d'une calotte sphrique tant gal  {@value}.
   */
  private static final double DEFAULT_CURVATURE = DEFAULT_R;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>raduis_of_curvature</b> correspond au rayon de courbure de la calotte sphrique.
   */
  private double radius_of_curvature;
  
  //--------------------------
  // PARAMTRE DE PRCALCUL //
  //--------------------------
  
  private SVector3d relative_sphere_position;
   
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * ...
   * 
   * @param position
   * @param normal
   * @param R
   * @param radius_of_curvature
   */
  public SSphericalCapGeometry(SVector3d position, SVector3d normal, double R, double radius_of_curvature) throws SConstructorException
  {
    this(position, normal, R, radius_of_curvature, null);
  }

  /**
   * ...
   * 
   * @param position
   * @param normal
   * @param R
   * @param parent
   * @throws SConstructorException
   */
  public SSphericalCapGeometry(SVector3d position, SVector3d normal, double R, double radius_of_curvature, SPrimitive parent) throws SConstructorException
  {
    super(position, normal, R, parent);
    
    // S'assurer que le rayon de courbure est plus grand que le rayon du disque de la calotte sphrique
    if(Math.abs(radius_of_curvature) < R)
      throw new SConstructorException("Erreur SSphericalCapGeometry 001 : Le rayon de courbure '" + radius_of_curvature + "' en valeur absolue doit tre plus lev que le rayon du disque tait '" + R + "'.");
    
    this.radius_of_curvature = radius_of_curvature;
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SSphericalCapGeometry 002 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }  
  }

  /**
   * ...
   * 
   * @param sbr
   * @param parent
   * @throws IOException
   * @throws SConstructorException
   */
  public SSphericalCapGeometry(SBufferedReader sbr, SPrimitive parent) throws IOException, SConstructorException
  {
    this(DEFAULT_POSITION, DEFAULT_SURFACE_NORMAL, DEFAULT_R, DEFAULT_CURVATURE, parent);
    
    //Faire la lecture
    try{
      read(sbr);    
    }catch(SInitializationException e){    
      throw new SConstructorException("Erreur SSphericalCapGeometry 003 : Une erreur est survenue lors de l'initialisation." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);  
    }
  
  }

  //------------
  // MTHODES //
  //------------
  
  @Override
  public int getCodeName()
  {
    return SAbstractGeometry.SPHERICAL_CAP_CODE;
  }
  
  /**
   * Mthode pour obtenir la position relative  cette calotte sphrique.
   * 
   * @return La position relative de la sphre  cette calotte sphrique.
   */
  public SVector3d getRelativeSpherePosition()
  {
    return relative_sphere_position;
  }
  
  /**
   * Mthode pour avoir l'paisseur de la calotte sphrique. 
   * Cette paisseur correspond  la distance entre le milieu de la calotte et le milieu du disque de la calotte.
   * @return L'paisseur de la calotte sphrique.
   */
  public double capSize()
  {
    return Math.abs(radius_of_curvature) - (relative_sphere_position.substract(position).modulus());
  }
  
  @Override
  public SRay intersection(SRay ray) throws SAlreadyIntersectedRayException
  {
    return ray;
  }
  
  @Override
  protected SVector3d evaluateIntersectionNormal(SRay ray, double intersection_t)
  {
    // valuation de la normale  la surface dans le sens extrieur  la courbure convexe de la sphre.
    return ray.getPosition(intersection_t).substract(relative_sphere_position).normalize();   
  }
  
  /**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
  {
    // S'assurer que le rayon de courbure est plus grand que le rayon du disque de la calotte sphrique
    if(Math.abs(radius_of_curvature) < R)
      throw new SRuntimeException("Erreur SSphericalCap 003 : Le rayon de courbure '" + radius_of_curvature + "' en valeur absolue doit tre plus lev que le rayon du disque tait '" + R + "'.");
    
    
    // valuons le dplacement par rapport  la position centrale du disque reprsentant 
    // le plan de la calotte sphrique  l'aide de pythagore.
    double displacement = Math.sqrt(radius_of_curvature*radius_of_curvature - R*R);
    
    // Si la courbure est positive, la courbure sera convexe (displacement < 0)
    // tel que propos dans la convension de signe des dioptres sphriques.
    // Le centre de la sphre relative dfinissant la courbure sera derrire la normale.
    
    // Si la courbure est ngative, la courbure sera concave (displacement > 0)
    // tel que propos dans la convension de signe des dioptres sphrique.
    // Le centre de la sphre relative dfinissant la courbure sera devant la normale.
    
    if(radius_of_curvature > 0)
      relative_sphere_position = position.substract(surface_normal.multiply(displacement));
    else
      relative_sphere_position = position.add(surface_normal.multiply(displacement));
  }
 
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_CURVATURE : radius_of_curvature = readDouble(remaining_line, SKeyWordDecoder.KW_CURVATURE); return true;
            
      default : return super.read(sbr, code, remaining_line);
    }
  }
  
  @Override
  public void write(BufferedWriter bw) throws IOException
  {
    bw.write(SKeyWordDecoder.KW_SPHERICAL_CAP);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    //crire les proprits de la classe SDiskGeometry et ses paramtres hrits
    writeSSphericalCapGeometryParameter(bw);
    
    bw.write(SKeyWordDecoder.KW_END);
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
  }

  /**
   * Mthode pour crire les paramtres associs  la classe SSphericalCapGeometry et ses paramtres hrits.
   * @param bw - Le BufferedWriter crivant l'information dans un fichier txt.
   * @throws IOException Si une erreur I/O s'est produite.
   * @see IOException
   */
  protected void writeSSphericalCapGeometryParameter(BufferedWriter bw)throws IOException
  {
    //crire les paramtres hrits de la classe SDiskGeometry
    writeSDiskGeometryParameter(bw);   
    
    bw.write(SKeyWordDecoder.KW_CURVATURE);
    bw.write("\t\t");
    bw.write(Double.toString(radius_of_curvature));
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  @Override
  protected SVectorUV evaluateIntersectionUV(SRay ray, double intersection_t)
  {
    throw new SNoImplementationException("Erreur SSpherialCapGeometry 005 : La mthode n'a pas encore t implmente.");
  }
  
  @Override
  protected void readingInitialization() throws SInitializationException
  {
    super.readingInitialization();
    
    initialize();
  }
  
  @Override
  public String getReadableName()
  {
    return SKeyWordDecoder.KW_SPHERICAL_CAP;
  }
  
  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}//fin de la classe SSphericalCapGeometry
