/**
 * 
 */
package sim.geometry;

import java.io.BufferedWriter;
import java.io.IOException;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.graphics.SPrimitive;
import sim.math.SAffineTransformation;
import sim.math.SMatrix4x4;
import sim.math.SVector3d;
import sim.math.SVectorUV;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SLog;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>STransformableGeometry</b> qui reprsente une gomtrie pouvant tre transforme  l'aide de matrices de transformations linaires
 * comme la <b>translation</b>, la <b>rotation</b> et <b>l'homothtie</b> (<i>scale</i>). Cette gomtrie devra contenir une gomtrie interne
 * donnant la forme de base (sans transformation)  la gomtrie transformable.
 * 
 * @author Simon Vzina
 * @since 2015-07-17
 * @version 2017-12-20 (version labo  Le ray tracer v2.1)
 */
public class STransformableGeometry extends SAbstractGeometry {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { 
    SKeyWordDecoder.KW_SCALE, SKeyWordDecoder.KW_ROTATION, SKeyWordDecoder.KW_TRANSLATION
  };
  
  /**
   * La constante <b>DEFAULT_GEOMETRY</b> correspond  la gomtrie par dfaut tant <b>null</b>.
   * Cela signifie qu'il n'y a pas de gomtrie interne  la gomtrie transformable.
   */
  private static final SGeometry DEFAULT_GEOMETRY = null;
  
  /**
   * La constante <b>DEFAULT_SCALE</b> correspond au vecteur comprenant les informations  attribuer  une matrice de transformation d'homothtie (<i>scale</i>).
   * Par dfaut, le vecteur sera (1.0, 1.0, 1.0) ce qui correspond  aucune dformation.
   */
  private static final SVector3d DEFAULT_SCALE = new SVector3d(1.0, 1.0, 1.0);
  
  /**
   * La constante <b>DEFAULT_ROTATION</b> correspond au vecteur comprenant les informations  attribuer  une matrice de transformation de rotation.
   * Par dfaut, le vecteur sera (0.0, 0.0, 0.0) ce qui correspond  aucune rotation.
   */
  private static final SVector3d DEFAULT_ROTATION = new SVector3d(0.0, 0.0, 0.0);
  
  /**
   * La constante <b>DEFAULT_TRANSLATION</b> correspond au vecteur comprenant les informations  attribuer  une matrice de transformation de translation.
   * Par dfaut, le vecteur sera (0.0, 0.0, 0.0) ce qui correspond  aucune translation.
   */
  private static final SVector3d DEFAULT_TRANSLATION = new SVector3d(0.0, 0.0, 0.0);
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>geometry</b> correspond  la gomtrie interne  la gomtrie transformable.
   */
  protected SGeometry geometry;       
  
  /**
   * La variable <b>scale</b> correspond au vecteur comprenant les informations  attribuer  une matrice de transformation d'homothtie (<i>scale</i>).
   */
  private SVector3d scale;              
  
  /**
   * La variable <b>rotation</b> correspond au vecteur comprenant les informations  attribuer  une matrice de transformation de rotation.
   */
  private SVector3d rotation;           
  
  /**
   * La variable <b>translation</b> correspond au vecteur comprenant les informations  attribuer  une matrice de transformation de translation.
   */
  private SVector3d translation;        
  
  /**
   * Constructeur d'une gomtrie transformable  l'aide d'une gomtrie interne.
   * @param geometry - La gomtrie interne.
   */
  public STransformableGeometry(SGeometry geometry)
  {
    this(geometry, DEFAULT_SCALE, DEFAULT_ROTATION, DEFAULT_TRANSLATION);
    
  }

  /**
   * Constructeur d'une gomtrie transformable  l'aide d'une gomtrie interne et des
   * matrices de transformations linaires comme <b>l'homoththie</b> (<i>scale</i>), <b>la rotation</b> et <b>la translation</b>.
   * @param geometry - La gomtrie transforme.
   * @param scale - Le vecteur dfinissant la matrice <b>d'homoththie</b> (<i>scale</i>).
   * @param rotation - Le vecteur dfinissant la matrice <b>de rotation</b> (en degr).
   * @param translate - Le vecteur dfinissant la matrice <b>de translation</b>.
   */
  public STransformableGeometry(SGeometry geometry, SVector3d scale, SVector3d rotation, SVector3d translation)
  {
    this(geometry, scale, rotation, translation, null);
  }
  
  /**
   * Constructeur d'une gomtrie transformable  l'aide d'une gomtrie interne reli  une primitive parent et des
   * matrices de transformations linaires comme <b>l'homoththie</b> (<i>scale</i>), <b>la rotation</b> et <b>la translation</b>.
   * @param geometry - La gomtrie transforme.
   * @param scale - Le vecteur dfinissant la matrice <b>d'homoththie</b> (<i>scale</i>).
   * @param rotation - Le vecteur dfinissant la matrice <b>de rotation</b> (en degr).
   * @param translate - Le vecteur dfinissant la matrice <b>de translation</b>.
   * @param parent - La primitive parent  cette gomtrie.
   * @throws SConstructorException Si une erreur est survenue lors de la construction de la gomtrie.
   */
  public STransformableGeometry(SGeometry geometry, SVector3d scale, SVector3d rotation, SVector3d translation, SPrimitive parent) throws SConstructorException
  {
    super(parent);
    
    this.geometry = geometry;
       
    this.scale = scale;
    this.rotation = rotation;
    this.translation = translation;
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur STransformableGeometry 001 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }  
  }
  
  /**
   * Constructeur d'une gomtrie transformable  partir d'information lue dans un fichier de format txt.
   * @param sbr - Le BufferedReader cherchant l'information dans le fichier txt.
   * @param parent - La primitive qui fait la construction de cette gomtrie (qui est le parent).
   * @throws IOException Si une erreur de l'objet SBufferedWriter est lance.
   * @throws SConstructorException Si une erreur est survenue lors de la construction de la gomtrie.
   * @see SBufferedReader
   */
  public STransformableGeometry(SBufferedReader sbr, SPrimitive parent) throws IOException, SConstructorException
  {
    this(DEFAULT_GEOMETRY, DEFAULT_SCALE, DEFAULT_ROTATION, DEFAULT_TRANSLATION, parent);   
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur STransformableGeometry 002 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }  
  }
  
  /**
   * Mthode pour obtenir la gomtrie interne  la gomtrie transformable.
   * 
   * @return La gomtrie interne.
   */
  public SGeometry getGeometry()
  {
    return geometry;
  }
  
  /**
   * Mthode pour obtenir le vecteur dfinissant l'homothtie (<i>scale</i>) de la gomtrie.
   * 
   * @return Le vecteur d'homothtie (<i>scale</i>).
   */
  public SVector3d getScale()
  {
    return scale;
  }
  
  /**
   * Mthode pour obtenir le vecteur dfinissant la rotation de la gomtrie.
   * 
   * @return Le vecteur de rotation.
   */
  public SVector3d getRotation()
  {
    return rotation;
  }
  
  /**
   * Mthode pour obtenir le vecteur dfinissant la translation de la gomtrie.
   * 
   * @return Le vecteur de translation.
   */
  public SVector3d getTranslation()
  {
    return translation;
  }
  
  @Override
  public int getCodeName()
  {
    return SAbstractGeometry.TRANSFORMABLE_CODE;
  }
  
  /**
   * Mthode qui dtermine s'il y a une gomtrie interne  la gomtrie transformable.
   * 
   * @return <b>true</b> s'il y a une gomtrie interne et <b>false</b> sinon.
   */
  private boolean isGeometrySelected(){
    return this.geometry != null;
  }
  
  @Override
  public boolean isClosedGeometry()
  {
    // S'il y a une gomtrie interne  la gomtrie transformable
    if(geometry != null)
      return geometry.isClosedGeometry();
    else
      return false;
  }

  @Override
  public boolean isInside(SVector3d v)
  {
	// S'il n'y a pas de gomtrie interne  la gomtrie transformable
	  if(geometry == null)
    	return false;
    		
	return false;
  }

  @Override
  public SRay intersection(SRay ray) throws SAlreadyIntersectedRayException
  {
    // Vrifier qu'il y a une gomtrie interne  la gomtrie transformable.
    if(geometry == null)
      return ray;
    
    // Modifiez  partir d'ici.
    return ray;
  }

  @Override
  public void write(BufferedWriter bw) throws IOException
  {
    bw.write(SKeyWordDecoder.KW_GEOMETRY);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    //crire les proprits de la classe SSphereGeometry et ses paramtres hrits
    writeSTransformableGeometryParameter(bw);
    
    bw.write(SKeyWordDecoder.KW_END);
    bw.write(SStringUtil.END_LINE_CARACTER);
    bw.write(SStringUtil.END_LINE_CARACTER);
  }

  /**
   * Mthode pour crire les paramtres associs  la classe STransformableGeometry et ses paramtres hrits.
   * @param bw - Le BufferedWriter crivant l'information dans un fichier txt.
   * @throws IOException Si une erreur I/O s'est produite.
   * @see IOException
   */
  protected void writeSTransformableGeometryParameter(BufferedWriter bw)throws IOException
  {
    bw.write(SKeyWordDecoder.KW_SCALE);
    bw.write("\t");
    scale.write(bw);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_ROTATION);
    bw.write("\t");
    rotation.write(bw);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_TRANSLATION);
    bw.write("\t");
    translation.write(bw);
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    if(geometry != null)
      geometry.write(bw);
    
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  @Override
  protected SVector3d evaluateIntersectionNormal(SRay ray, double intersection_t)
  {
    throw new SNoImplementationException("Erreur STransformableGeometry 005 : Cette mthode n'est pas implmente, car elle n'est pas utile au fonctionnement de la classe.");
  }

  /**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
  {
    
  }

  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException, IOException
  {
    //Analyser le code sur la lecture d'une gomtrie  l'aide du lecteur de gomtrie
    SGeometryReader reader = new SGeometryReader(sbr, code, null);
    
    //S'il y a eu lecture d'une gomtrie
    if(reader.isRead())
      if(!isGeometrySelected())
      {
        geometry = reader.getGeometry();
        return true;
      }
      else
        throw new SReadingException("Erreur STranformableGeometry 006 : Une gomtrie a dj t slectionne pour cette gomtrie transformable.");   
      
    //Analyser le code sur d'autres paramtres
    switch(code)
    {
      case SKeyWordDecoder.CODE_SCALE : scale = new SVector3d(remaining_line); return true;
      
      case SKeyWordDecoder.CODE_ROTATION : rotation = new SVector3d(remaining_line); return true;
      
      case SKeyWordDecoder.CODE_TRANSLATION : translation = new SVector3d(remaining_line); return true;
      
      default : return false;
    }
  }

  @Override
  protected SVectorUV evaluateIntersectionUV(SRay ray, double intersection_t)
  {
    throw new SNoImplementationException("Erreur STransformableGeometry 007 : La mthode n'a pas t implmente.");
  }
  
  @Override
  protected void readingInitialization() throws SInitializationException
  {
    super.readingInitialization();
    
    initialize();
    
    //Affichage de message si la lecture est incomplte
    if(!isGeometrySelected())
      SLog.logWriteLine("Message STransformableGeometry : Une gomtrie transformable (Geometry) a t lue mais initialise sans gomtrie interne.");
  }
  
  @Override
  public String getReadableName()
  {
    return SKeyWordDecoder.KW_GEOMETRY;
  }
  
  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    // Mettre les mots cl du lecteur de gomtrie (SGeometryReader)
    other_parameters = SStringUtil.merge(other_parameters, SGeometryReader.KEYWORD_PARAMETER);
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}//fin de la classe STransformableGeometry
