/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.geometry.SGeometry;
import sim.math.SStatistical;
import sim.math.SVector3d;

/**
 * <p>
 * La classe <b>SBoundingBox</b> reprsente une bote englobant une gomtrie align sur les axes x,y et z.
 * Cette bote contient une rfrence  une gomtrie  l'intrieur de celle-ci.
 * </p>
 * 
 * <p>
 * Puisque la classe <b>SBoundingBox</b> implmente la classe <b>Comparable</b>, elle peut tre
 * trie en fonction de sa taille maximale dtermine par la mthode getMaxLenght().
 * </p>
 * 
 * <p>
 * Puisque la classe <b>SBoundingBox</b> implmente la classe <b>SStatistical</b>, elle peut tre
 * analyse statistiquement en fonction de sa taille maximale dtermine par la mthode getMaxLenght().
 * </p>
 * 
 * @author Simon Vzina
 * @since 2015-08-04
 * @version 2016-02-07
 */
public class SBoundingBox implements Comparable<SBoundingBox>, SStatistical {

  /**
   * La variable <b>min_point</b> reprsente la coordonne (x,y,z) <b>minimale</b> de la bote.
   * En comparaison avec un carr  deux dimensions, cela reprsenterait le coin infrieur gauche.
   */
  private final SVector3d min_point;    
  
  /**
   * La variable <b>max_point</b> reprsente la coordonne (x,y,z) <b>maximale</b> de la bote.
   * En comparaison avec un carr  deux dimensions, cela reprsenterait le coin suprieur droit.
   */
  private final SVector3d max_point;    //la coordonne (x,y,z) maximale de la bote
  
  /**
   * La variable <b>geometry</b> correspond  la gomtrie  l'intrieur de la bote englobante.
   */
  private final SGeometry geometry;     
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'une bote englobante  partir de deux points extrmums.
   * 
   * @param geometry - La gomtrie  l'intrieur de la bote.
   * @param min_point - Le point minimal de la bote.
   * @param max_point - Le point maximal de la bote.
   * @throws SConstructorException Si min_point n'est pas totalement infrieur au max_point.
   */
  public SBoundingBox(SGeometry geometry, SVector3d min_point, SVector3d max_point) throws SConstructorException
  {
    if(min_point.getX() > max_point.getX() || min_point.getY() > max_point.getY() || min_point.getZ() > max_point.getZ())
      throw new SConstructorException("Erreur SBoundingBox 001 : Le point min " + min_point.toString() + " n'est pas infrieur au point max " + max_point.toString() + ".");
    
    this.geometry = geometry;
    this.min_point = min_point;
    this.max_point = max_point;
  }

  /**
   * Constructeur d'une bote englobant  partir de plusieurs point pouvant dfinir les extrmits de la bote englobante.
   * 
   * @param geometry La gomtrie  l'intrieur de la bote. 
   * @param point_list La liste des points.
   * @throws SConstructorException Si la liste contient moins de deux points.
   */
  public SBoundingBox(SGeometry geometry, List<SVector3d> point_list) throws SConstructorException
  {
    if(point_list.size() < 2)
      throw new SConstructorException("Erreur SBoundingBox 002 : La list contient " + point_list.size() + " points ce qui est infrieur  2.");
    
    this.geometry = geometry;
    
    min_point = SVector3d.findMinValue(point_list);
    max_point = SVector3d.findMaxValue(point_list);
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir la gomtrie  l'intrieur de la bote englobante.
   * 
   * @return La gomtrie  l'intrieur de la bote englobante.
   */
  public SGeometry getGeometry()
  {
    return geometry;
  }
  
  /**
   * Mthode pour obtenir la coordonne minimale de la bote englobante.
   * 
   * @return Le point minimal de la bote englobante.
   */
  public SVector3d getMinPoint()
  {
    return min_point;
  }
  
  /**
   * Mthode pour obtenir la coordonne maximale de la bote englobante.
   * 
   * @return Le point maximal de la bote englobante.
   */
  public SVector3d getMaxPoint()
  {
    return max_point;
  }
  
  /**
   * Mthode pour obtenir la liste des huit points reprsentant la bote englobante.
   * 
   * @return La liste des points. 
   */
  public List<SVector3d> getListPoint()
  {
    List<SVector3d> list = new ArrayList<SVector3d>();
    
    // Ajouter les points dans le plan min_z
    list.add(min_point);
    list.add(new SVector3d(max_point.getX(), min_point.getY(), min_point.getZ()));
    list.add(new SVector3d(max_point.getX(), max_point.getY(), min_point.getZ()));
    list.add(new SVector3d(min_point.getX(), max_point.getY(), min_point.getZ()));
    
    // Ajouter les points dans le plan max_z
    list.add(new SVector3d(min_point.getX(), min_point.getY(), max_point.getZ()));
    list.add(new SVector3d(max_point.getX(), min_point.getY(), max_point.getZ()));
    list.add(max_point);
    list.add(new SVector3d(min_point.getX(), max_point.getY(), max_point.getZ()));
    
    return list;
  }
  
  /**
   * Mthode pour obtenir le volume de la bote englobante.
   * Puisque la bote est aligne sur l'axe x,y et z, l'quation du volume V correspond 
   * <ul> V = dx * dy * dz</ul>
   * <p>
   * Il y a cependant une mise en garde avec cette mthode. Si une gomtrie est planaire (comme un triangle) et qu'elle
   * est situe dans l'un des 6 plans des axes x,y et z, cette mthode retournera un volume trs prs de <b>zro</b>.
   * </p>
   * 
   * @return Le volume de la bote englobante.
   */
  public double getVolume()
  {
    return (max_point.getX() - min_point.getX()) * (max_point.getY() - min_point.getY()) * (max_point.getZ() - min_point.getZ());
  }
  
  /**
   * Mthode pour obtenir la longueur moyenne de la bote englobante.
   * Puisque la bote est aligne sur l'axe x,y et z, l'quation de la longueur moyenne correspond 
   * <ul> L_moy = ( dx + dy + dz ) / 3</ul>
   * 
   * @return La longueur moyenne de la bote.
   */
  public double getAverageLenght()
  {
    return ( (max_point.getX() - min_point.getX()) + (max_point.getY() - min_point.getY()) + (max_point.getZ() - min_point.getZ()) ) / 3;
  }
  
  /**
   * Mthode pour obtenir la dimension la plus grande de la bote englobante.
   * Puisuqe la bote est aligne sur l'axe x, y et z, l'quation de la dimension la plus grande correspond 
   * <ul> L_max = max { dx, dy, dz }</ul>
   * 
   * @return La dimension la plus grande de la bote.
   */
  public double getMaxLenght()
  {
    double dx = max_point.getX() - min_point.getX();
    double dy = max_point.getY() - min_point.getY();
    double dz = max_point.getZ() - min_point.getZ();
    
    double max = dx;
    
    if(max < dy)
      max = dy;
    
    if(max < dz)
      max = dz;
    
    return max;
  }

  @Override
  public int compareTo(SBoundingBox box)
  {
    return Double.compare(this.getMaxLenght(), box.getMaxLenght());
  }

  @Override
  public double getStatisticalValue()
  {
    return getMaxLenght();
  }
  
}//fin de la classe SBoundingBox
