/**
 * 
 */
package sim.graphics;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;
import sim.geometry.SAbstractGeometry;
import sim.geometry.SBTriangleGeometry;
import sim.geometry.SGeometry;
import sim.geometry.STransformableGeometry;
import sim.geometry.STriangleGeometry;
import sim.graphics.material.SMaterial;
import sim.graphics.material.STextureMaterial;
import sim.loader.SLoaderException;
import sim.loader.model.SModelLoader;
import sim.math.SImpossibleNormalizationException;
import sim.math.SVector3d;
import sim.readwrite.SAbstractReadable;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SLog;
import sim.util.SReader;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SModelReader</b> reprsente un lecteur de modle 3d. 
 * L'objectif de ce lecteur sera la lecture des proprits du modle dans le fichier de scne comme le nom du fichier principalement.
 * 
 * @author Simon Vzina
 * @since 2015-07-22
 * @version 2020-01-08 (Version : Le ray tracer v2.107)
 */
public class SModelReader extends SAbstractReadable implements SReader {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { 
    SKeyWordDecoder.KW_FILE, SKeyWordDecoder.KW_SCALE, SKeyWordDecoder.KW_ROTATION,
    SKeyWordDecoder.KW_TRANSLATION, SKeyWordDecoder.KW_UV_FORMAT  
  };
  
  /**
   * La constante <b>DEFAULT_MODEL</b> correspond au modle par dfaut tant inexistant.
   * Sa valeur est alors <b>null</b>.
   */
  private static final SModel DEFAULT_MODEL = null;    //modle par dfaut tant non charg
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>model_map</b> correspond  la carte des modles dj lu et charg en mmoire.
   * On peut ainsi les rutiliser et leur appliquer de nouvelles transformation pour les visualiser diffremment.
   * La <b>cl de recherche</b> est le <b>nom du fichier</b>.
   */
  private static final Map<String, SModel> model_map = new HashMap<String, SModel>();
  
  /**
   * La variable <b>file_name</b> correspond au nom du fichier comprenant les informations dfinissant le modle.
   */
  private String file_name;                
  
  /**
   * La variable <b>scale</b> correspond au vecteur comprenant les informations de la transformation d'homthtie (<i>scale</i>).
   */
  private SVector3d scale;                  
  
  /**
   * La variable <b>rotation</b> correspond au vecteur comprenant les informations de la transformation de rotation.
   */
  private SVector3d rotation;               
  
  /**
   * La variable <b>translation</b> correspond au vecteur comprenant les informations de la transformation de translation.
   */
  private SVector3d translation;           
  
  /**
   * La variable <b>uv_format</b> correspond au code d'interprtation des coordonnes uv que l'on doit appliquer au texture
   * afin que chaque coordonne uv corresponde au bon texel des textures.
   */
  private int uv_format;
  
  /**
   * La variable <b>model</b> correspond au modle lu par le lecteur.
   */
  SModel model;
  
  /**
   * La varaible <b>is_read</b> dtermine si un modle a t lu avec succs par le lecteur.
   */
  boolean is_read;
  
  //--------------------------------------------------
  //Paramtres de prcalculs pour les transformations
  //--------------------------------------------------
  
    
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Constructeur d'un lecteur de modle par dfaut. 
   */
  public SModelReader()
  {
    file_name = SModel.DEFAULT_FILE_NAME;
    
    scale = SModel.DEFAULT_SCALE;
    rotation = SModel.DEFAULT_ROTATION;
    translation = SModel.DEFAULT_TRANSLATION;
    
    uv_format = STexture.UV_DEFAULT;
    
    model = DEFAULT_MODEL;
    is_read = false;
  }

  /**
   * Constructeur d'un constructeur de gomtrie  partir d'information lue dans un fichier de format txt.
   * @param sbr - Le BufferedReader cherchant l'information dans le fichier txt.
   * @throws IOException Si une erreur de l'objet SBufferedWriter est lance.
   * @throws SConstructorException Si une erreur est survenue lors de la construction.
   */
  public SModelReader(SBufferedReader sbr)throws IOException, SConstructorException
  {
    this();
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SModelReader 001 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }  
  }
  
  @Override
  public SModel getValue() throws SRuntimeException
  {
    if(is_read)
      return model;
    else
      throw new SRuntimeException("Erreur SModelReader 002 : Le modle '" + file_name + "' n'a pas t lu.");
  }

  @Override
  public boolean asRead()
  {
    return is_read;
  }
    
  /**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
  {
    
     try{
      
      //Vrifier si le modle a dj t lu
      if(model_map.containsKey(file_name))
        model = model_map.get(file_name);
      else
      {
        SModelLoader model_loader = new SModelLoader();
      
        SLog.logWriteLine("Message SModelReader : Lecture du modle '" + file_name + "'.");
        
        model = model_loader.loadModel(file_name);  //lecture du modle (exception lance s'il y a eu erreur)
        
        model_map.put(file_name, model);            //mettre le modle lu dans la carte des modles
      }
      
      // Excuter la transformation du modle par la construction d'un nouveau.
      model = transformModel(model);
      
      is_read = true;
      
    }catch(SLoaderException e){
      SLog.logWriteLine("Erreur SModelReader 003 : Le chargement du modle '" + file_name + "' est impossible. " + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
    }
    
       
  }

  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException, IOException
  {
    //Lecture des diffrentes gomtrie reconnues par le lecteur
    switch(code)
    {
      case SKeyWordDecoder.CODE_FILE : file_name = readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_FILE); return true;
    
      case SKeyWordDecoder.CODE_SCALE : scale = new SVector3d(remaining_line); return true;
      
      case SKeyWordDecoder.CODE_ROTATION : rotation = new SVector3d(remaining_line); return true;
      
      case SKeyWordDecoder.CODE_TRANSLATION : translation = new SVector3d(remaining_line); return true;
      
      case SKeyWordDecoder.CODE_UV_FORMAT : uv_format = readIntOrExpression(remaining_line, SKeyWordDecoder.KW_UV_FORMAT, STexture.UV_FORMAT); return true;
      
      default : return false;
    }
  }

  /**
   * Mthode pour transformer l'intgralit des gomtries contenues dans le modle en fonction des matrices de transformation.
   * Un nouveau modle avec l'applications des matrices de transformation sera gnr.
   * 
   * @param model Le modle  transformer.
   * @return Un nouveau modle o l'application des matrices de transformation a t ralise.
   */
  private SModel transformModel(SModel model)
  {
	  // Construction d'un nouveau modle
    SModel transformed_model = new SModel(file_name, scale, rotation, translation, uv_format);
    
    // Obtenir la liste des primitives du modle
    List<SPrimitive> list = model.getPrimitiveList();
    
    // Paramtre dterminant si une transformation de la gomtrie de la primitive est ncessaire
    boolean transformation_required = !(scale.equals(SModel.DEFAULT_SCALE) && rotation.equals(SModel.DEFAULT_ROTATION) && translation.equals(SModel.DEFAULT_TRANSLATION));
    
    if(!transformation_required)
      SLog.logWriteLine("Message SModelReader : Le modle " + file_name + " sera charg sans transformation.");
    
    // Gomtrie en erreur de transformation
    int geometry_error = 0;                     
      
    // Itrer sur l'ensemble des gomtries
    for(SPrimitive p : list)
    {
      SGeometry geometry;                       // la primitive transforme  ajouter  la liste
      SMaterial material = p.getMaterial();     // le matriel de la gomtrie  
        
      // Modification des gomtries en fonction du type qu'elle est
      // Des erreurs sont possibles si la transformation est impossible.
      // Par exemple : Un modle peut tre mal dfini pour certains triangles.
      try{
          
          // Vrifier si la transformation de la gomtrie est ncessaire.
          if(transformation_required)
          {
            switch(p.getGeometry().getCodeName())
            {
              case SAbstractGeometry.TRIANGLE_CODE :  geometry = transformTriangleGeometry((STriangleGeometry)p.getGeometry()); break;
              
              case SAbstractGeometry.BTRIANGLE_CODE : geometry = transformBTriangleGeometry((SBTriangleGeometry)p.getGeometry()); break;
              
              default :                               geometry = new STransformableGeometry(p.getGeometry(), scale, rotation, translation); break;
            }
          }
          else
            geometry = p.getGeometry();   // gomtrie non transforme
          
          // Modifier le format d'interprtation des coordonnes uv de texture pour un matriel avec texture.
          // Cependant, plusieurs instance du modle peuvent tre construite.
          // Une seule interprtation des coordonne uv sera possible. Ce sera la premire dfinition qui sera retenue.
          if(material.asTexture())
          {
            STextureMaterial texture_material = (STextureMaterial) material;
            
            if(!texture_material.isUVFormatSelected())
              texture_material.setUVFormat(uv_format);
          }
          
          // Ajouter la nouvelle gomtrie dans une nouvelle primitive et l'injecter dans le modle. 
          if(transformation_required)
            transformed_model.addPrimitive(new SPrimitive(geometry, material));      
          else
            transformed_model.addPrimitive(p);  // remettre l'ancienne primitive dans la nouveau modle    
          
        }catch(SConstructorException e){
          //S'il y a des gomtries en erreur de construction aprs l'application des transformations
          geometry_error++;
        }catch(SImpossibleNormalizationException e){
          geometry_error++;
        }   
      }//fin for
        
    //Afficher un message sur le nombre de gomtrie en erreur de transformation
    if(geometry_error > 0)
      SLog.logWriteLine("Message SModelReader : Il y a '" + geometry_error + "' gomtries en erreur de transformation. Ils ne seront pas disponibles pour l'affichage.");
  	 
    return transformed_model;
  }
  
  /**
   * Mthode pour faire la transformation d'un STriangleGeometry dans l'espace objet (unitaire) vers l'espace monde (transform)  partir des vecteurs de transformation lus par le lecteur.
   * 
   * @param triangle Le triangle  transform
   * @return La gomtrie d'un triangle transform.
   */
  private SGeometry transformTriangleGeometry(STriangleGeometry triangle)
  {
    //-----------------------------------------------------------------------------------
    // POUR LABORATOIRE :
    // Vous devrez retourner un nouveau triangle transform.
    // Par example, un nouveau triangle "non transform" correspondrait 
    //  l'instruction suivante :
    // return new STriangleGeometry(triangle.getP0(), triangle.getP1(), triangle.getP2());
    //------------------------------------------------------------------------------------
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour faire la transformation d'un SBTriangleGeometry dans l'espace objet (unitaire) vers l'espace monde (transform)  partir des vecteurs de transformation lus par le lecteur.
   * 
   * @param triangle Le triangle  transform.
   * @return La gomtrie d'un triangle barycentrique transform.
   */
  private SGeometry transformBTriangleGeometry(SBTriangleGeometry triangle) 
  {
    //-----------------------------------------------------------------------------------------
    //POUR LABORATOIRE :
    // Vous devrez retourner un nouveau triangle barycentrique transform.
    // Par example, un nouveau triangle barycentrique "non transform" correspondrait 
    //  l'instruction suivante :
    // return new SBTriangleGeometry(triangle.getP0(), triangle.getP1(), triangle.getP2(),
    //                               triangle.getN0(), triangle.getN1(), triangle.getN2(), 
    //                               triangle.getUV0(), triangle.getUV1(), triangle.getUV2());
    //-----------------------------------------------------------------------------------------
    throw new SNoImplementationException();
   
  }
  
  @Override
  protected void readingInitialization() throws SInitializationException
  {
    initialize();
  }
  
  @Override
  public String getReadableName()
  {
    return SKeyWordDecoder.KW_MODEL;
  }
  
  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
