/**
 * 
 */
package sim.graphics;

import java.io.BufferedWriter;
import java.io.IOException;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.geometry.SGeometry;
import sim.geometry.SGeometryReader;
import sim.graphics.material.SDefaultMaterial;
import sim.graphics.material.SMaterial;
import sim.readwrite.SAbstractReadableWriteable;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SLog;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SPrimitive</b> reprsente un objet comprenant une gomtrie et un matriel.
 * 
 * @author Simon Vzina
 * @since 2015-01-06
 * @version 2016-05-20
 */
public class SPrimitive extends SAbstractReadableWriteable{

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { SKeyWordDecoder.KW_MATERIAL_NAME };
  
  //-------------
  // VARIABLES //
  //-------------
  
	private SGeometry geometry;				    //la gomtrie de la primitive
	private SMaterial material;				    //le matriel de la primitive
	
	private String material_name;			    //nom du matriel si affectation  l'extrieur 
	private boolean isGeometry_selected;	//dtermine si la gomtrie a t slectionn lors de la lecture
	
	/**
	 * Constructeur d'une primitive lors d'une lecture d'un fichier de chargement.
	 */
	private SPrimitive()
	{
	  this(null, null);
	}
	
	/**
	 * Constructeur d'une primitive sans matriel.
	 * 
	 * @param geometry La gomtrie affecte  la primitive.
	 */
	public SPrimitive(SGeometry geometry)
	{
	  this(geometry, null);
	}
	
	/**
	 * Constructeur avec une gomtrie et un matriel sans relation pralable.
	 * @param geometry - La gomtrie de la primitive.
	 * @param material - Le matriel de la primitive.
	 * @throws SConstructorException Si une erreur est survenue lors de la construction.
	 */
	public SPrimitive(SGeometry geometry, SMaterial material) throws SConstructorException
	{
	  // Affectation de la gomtrie
	  this.geometry = geometry;
	  
	  if(geometry == null)
	    isGeometry_selected = false;
	  else
	  {
	    isGeometry_selected = true;
	    geometry.setPrimitiveParent(this);
	  }
	  
	  // Affectation du matriel
	  this.material = material;
	  
	  if(material == null)
	    material_name = SDefaultMaterial.DEFAULT_MATERIAL_NAME;
	  else
	    material_name = material.getName();
	  
	  try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SPrimitive 001 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }
	}
	
	/**
	 * Constructeur d'une primitive  partir d'information lue dans un fichier de format txt.
	 * @param sbr - Le BufferedReader cherchant l'information dans le fichier txt.
	 * @throws IOException Si une erreur de l'objet SBufferedWriter est lance.
	 * @throws SConstructorException Si une erreur est survenue  la construction.
	 * @see SBufferedReader
	 */
	public SPrimitive(SBufferedReader sbr)throws IOException, SConstructorException
	{
		this();		
		
		try{
		  read(sbr);
		}catch(SInitializationException e){
      throw new SConstructorException("Erreur SPrimitive 002 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }  
	}
	
	/**
	 * Mthode pour obtenir le nom du matriel  attribuer  cette primitive.
	 * @return Le nom de du matriel.
	 */
	public String getMaterialName()
	{ 
	  return material_name;
	}
	
	/**
	 * Mthode pour obtenir la gomtrie de la primitive. 
	 * Il est important de prciser qu'une primitive <b>sans gomtrie</b> retourne <b>null</b> comme rfrence.
	 * 
	 * @return La gomtrie de la primitive ou <b>null</b> si elle n'en a pas.
	 */
	public SGeometry getGeometry()
	{ 
	  return geometry;
	}
	
	/**
	 * Mthode pour obtenir le matriel affect  la primitive. 
	 * @return Le matriel de la primitive.
	 * @throws SRuntimeException Si l'affectation du matriel  la primitive n'a pas t encore effectue. 
	 */
	public SMaterial getMaterial()throws SRuntimeException
	{ 
		if(material == null)
			throw new SRuntimeException("Erreur SPrimitive 004 : L'affectation du matriel  cette primitive n'a pas encore t ralise.");
		else
			return material;
	}
	
	/**
	 * Mthode pour affecter un matriel  une primitive. L'affectation sera ralise si le nom du matriel est celui dfini dans la primitive.
	 * @param new_material - Le matriel  affecter  la primitive.
	 * @throws SRuntimeException Si le nom du matriel n'est pas celui accept par la primitive.
	 */
	public void setMaterial(SMaterial new_material)throws SRuntimeException
	{
		//Vrification que le nom du matriel accept par cette primitive est bien celui du nouveau matriel en paramtre
		if(new_material.getName().equals(material_name))
			material = new_material;
		else
			throw new SRuntimeException("Erreur SPrimitive 005 : L'affectation de ce matriel n'est pas possible puisque son nom n'est pas celui accept par cette primitive.");
	}
	
	@Override
	public void write(BufferedWriter bw) throws IOException 
	{
		bw.write(SKeyWordDecoder.KW_PRIMITIVE);
		bw.write(SStringUtil.END_LINE_CARACTER);
		
		//crire les paramtres de la classe SPrimitive
		writeSPrimitiveParameter(bw);
		
		bw.write(SKeyWordDecoder.KW_END);
		bw.write("\t#end primitive" + SStringUtil.END_LINE_CARACTER);
		bw.write(SStringUtil.END_LINE_CARACTER);
	}

	/**
	 * Mthode pour crire les paramtres associs  la classe SPrimitive.
	 * @param bw - Le BufferedWriter crivant l'information dans un fichier txt.
	 * @throws IOException S'il y a une erreur d'criture.
	 * @see IOException
	 */
	protected void writeSPrimitiveParameter(BufferedWriter bw)throws IOException 
	{
		bw.write(SKeyWordDecoder.KW_MATERIAL_NAME);
		bw.write("\t");
		bw.write(material_name);
		bw.write(SStringUtil.END_LINE_CARACTER);
		
		if(geometry != null)
			geometry.write(bw);
	}
	
	/**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
	{
		
	}

	/* (non-Javadoc)
	 * @see simTools.SReadableWriteable#read(SBufferedReader sbr, int code, String remaining_line) throws Exception
	 */
	@Override
	protected boolean read(SBufferedReader sbr, int code, String remaining_line)throws SReadingException, IOException
	{
		//Analyser le code sur la lecture d'une gomtrie  l'aide du lecteur de gomtrie
	  SGeometryReader reader = new SGeometryReader(sbr, code, this);
    
	  //S'il y a eu lecture d'une gomtrie
    if(reader.isRead())
      if(!isGeometry_selected)
      {
        geometry = reader.getGeometry();
        isGeometry_selected = true;
        return true;
      }
      else
        throw new SReadingException("Erreur SPrimitive 006 : Une gomtrie a dj t slectionne pour cette primitive.");   
      
	  //Analyser le code sur d'autres paramtres
	  switch(code)
		{
			case SKeyWordDecoder.CODE_MATERIAL_NAME : material_name = readStringNotEmpty(remaining_line, "nom du matriel"); return true;
			
			default : return false;
		}
	}
	
	@Override
	protected void readingInitialization() throws SInitializationException
	{
	  initialize();
	  
	  //Affichage de message si la lecture est incomplte
    if(!isGeometry_selected)
      SLog.logWriteLine("Message SPrimitive : Une primitive a t initialise sans gomtrie aprs une lecture.");
    
    if(material_name.equals(SDefaultMaterial.DEFAULT_MATERIAL_NAME))
      SLog.logWriteLine("Message SPrimitive : Une primitive a t initialise sans nom de matriel aprs une lecture.");
	}
	 
	@Override
  public String getReadableName()
  {
    return SKeyWordDecoder.KW_PRIMITIVE;
  }
  
  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    // Mettre les mots cl du lecteur de gomtrie (SGeometryReader)
    other_parameters = SStringUtil.merge(other_parameters, SGeometryReader.KEYWORD_PARAMETER);
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}//fin classe SPrimitive
