/**
 * 
 */
package sim.graphics.shader;

import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.geometry.SRay;
import sim.geometry.space.SGeometrySpace;
import sim.graphics.SColor;
import sim.graphics.light.SLight;
import sim.graphics.material.SMaterial;
import sim.math.SVector3d;
import sim.physics.SGeometricalOptics;

/**
 * La classe <b>SRecursiveShader</b> reprsente un shader avec lanc de rayon rcursif.
 * 
 * @author Simon Vzina
 * @since 2015-02-03
 * @version 2017-06-04
 */
public class SRecursiveShader extends SPhongReflexionShader {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>MINIMAL_MAX_DEPT</b> correspond au niveau de rcursivit minimale tant gal  {@value}.
   */
	private static final int MINIMAL_MAX_DEPT = 1; 
	
	//-------------
	// VARIABLES //
	//-------------
	
	/**
	 * La variable <b>max_dept</b> correspond au niveau de rcursivit qui sera utilis lors des calculs d'illuminations.
	 */
  private final int max_dept;	
	
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
	/**
	 * Constructeur d'un shader avec lanc de rayon rcursif.
	 * 
	 * @param geometry_space L'espace des gomtries.
	 * @param t_max Le temps de dplacement maximal d'un rayon.
	 * @param light_list La liste des sources de lumires.
	 * @param reflexion_algo Le type d'algorithme pour raliser le calcul de la rflexion.
	 * @param max_dept - Le niveau de rayon rcursif.
	 * @throws SConstructorException Si le niveau de rcursivit des rayons est infrieur au seuil minimal (habituellement 1).
   */
	public SRecursiveShader(SGeometrySpace geometry_space, double t_max, List<SLight> light_list, int reflexion_algo, int max_dept)throws SConstructorException 
	{
		super(geometry_space, t_max, light_list, reflexion_algo);
		
		if(max_dept < MINIMAL_MAX_DEPT)
      throw new SConstructorException("Erreur SRecursiveShader 002 : Le niveau '" + max_dept + "' de rcursivit des rayons doit tre suprieur  '" + MINIMAL_MAX_DEPT + "'.");
    
		this.max_dept = max_dept;
	}

	//------------
	// MTHODES //
	//------------
	
	@Override
	public SColor shade(SRay ray) throws SRuntimeException 
	{
		if(ray.asIntersected())
			throw new SRuntimeException("Erreur SRecursiveShader 003 : Le rayon a dj intersect une gomtrie pralablement.");
		    
		//Lancer un rayon de niveau de rcursivit 1
		return recursiveShade(ray, 1);
	}

	/**
	 * Mthode qui effectue l'illumination d'un rayon rcursivement.
	 * 
	 * @param ray Le rayon  illuminer.
	 * @param depth Le niveau de rcursivit du rayon.
	 * @return La couleur associe  l'illumination du rayon rcursif.
	 */
	private SColor recursiveShade(SRay ray, int depth)
	{
		// Si le niveau de rayon rcursif est trop profond, tout arrter et retourner la couleur noire
		if(depth > max_dept)
			return SIllumination.NO_ILLUMINATION;
		
		// Effectuer l'intersection avec l'espace des gomtries
		ray = geometry_space.nearestIntersection(ray, t_max);
			
		// S'il n'y a pas d'intersection, la couleur affecte sera noire
		if(!ray.asIntersected())
			return SIllumination.NO_ILLUMINATION;
		
		// Obtenir le matriel de la primitive
    SMaterial material = ray.getGeometry().getPrimitiveParent().getMaterial(); 
    
		// Vrifier s'il y a un algorithme d'illumination (pas en mode no_light)
    if(reflexion_algo == NO_LIGHT)
      if(ray.asUV())
       return material.diffuseColor(ray.getUV()); 
     else
       return material.diffuseColor();  
   
		// Constribution de l'illumination directe
		SColor color = directIllumination(ray, material);
			
		// Contribution de l'illumination indirecte (avec rcursivit potentielle)
		color = color.add(indirectIllumination(ray, material, depth));
			
		return color;
	}
	
	/**
	 * Mthode permettant d'valuer la couleur attribue  un rayon par illumination indirecte.
	 * @param ray Le rayon ayant ralis une intersection avec une gomtrie de l'espace des gomtries.
	 * @param material Le matriel appliqu sur la gomtrie intersect par le rayon.
	 * @param depth Le niveau de rcursivit du rayon.
	 * @return La couleur de l'illumination indirecte.
	 */
	protected SColor indirectIllumination(SRay ray, SMaterial material, int depth)
	{
	  // La somme des couleurs par rflexion et rfraction
	  SColor color = SIllumination.NO_ILLUMINATION;
	  
	  //-----------------------------------------------
    //  Illumination indirecte : Loi de la rflexion  
    //-----------------------------------------------
    if(material.isReflective())
    {
      // valuer la nouvelle orientation du rayon selon la loi de la rflexion.
      SVector3d R = SGeometricalOptics.reflexion(ray.getDirection(), ray.getShadingNormal());
    
      // Rayon de rflexion, sans changement de milieu (indice de rfraction du rayon intersect)
      SRay reflexion_ray = ray.castRecursiveRay(R, ray.getRefractiveIndex());
            
      // Ajouter la luminosit rflexive  la somme des couleurs en augmentant le niveau de 1
      color = color.add(recursiveShade(reflexion_ray, depth+1).multiply(material.reflectivity()));
    }
      
    //------------------------------------------------
    //  Illumination indirecte : Loi de la rfraction
    //------------------------------------------------
    if(material.isTransparent())
    {
      // Rfraction uniquement si la gomtrie intersecte est une gomtrie ferme (pas une surface)
      if(ray.getGeometry().isClosedGeometry())
      {
        // Indice de rfraction  dterminer (n1 = incident, n2 = rfract)
        double n1 = ray.getRefractiveIndex();
          
        //Pour dterminer n2, il y a deux scnarios  considrer :
        //1) L'intersection vient de l'extrieur, donc n2 sera l'indice interne de la gomtrie intersecte.
        //2) L'intersection vient de l'intrieur, donc on utilise le shader pour dterminer l'indice  l'extrieur de la gomtrie intersecte de l'intrieur.
          double n2;
          
          boolean from_outside;
          
          if(!ray.isInsideIntersection())                                  //de l'extrieur
          {
            n2 = material.refractiveIndex();
            from_outside = true;
          }
          else                                                             //de l'intrieur
          {
            n2 = evaluateRefractiveIndex(ray.getIntersectionPosition());
            from_outside = false;
          }
          
          // Vrifier qu'il n'y a pas rflexion totale interne
          if(!SGeometricalOptics.isTotalInternalReflection(ray.getDirection(), ray.getShadingNormal(), n1, n2))
          {
            // valuer l'orientation du rayon selon la loi de la rfraction.
            SVector3d T = SGeometricalOptics.refraction(ray.getDirection(), ray.getShadingNormal(), n1, n2);
          
            // Rayon de rfraction avec changement de milieu (indice de rfraction n2)
            SRay refraction_ray = ray.castRecursiveRay(T, n2);
            
            // Ajouter la luminosit rfractive  la somme des couleurs en augmentant le niveau de 1
            // Venant de l'extrieur, nous allons appliquer un filtrage  la couleur
            if(from_outside)
              color = color.add(recursiveShade(refraction_ray, depth+1).multiply(material.transparencyColor()));  // couleur avec filtrage
            else  
              color = color.add(recursiveShade(refraction_ray, depth+1).multiply(material.transparency()));       // couleur dj filtre
          }
        }
      }
    
    return color;
	}
	
}//fin classe SRecursiveShader
